import uuid
from collections.abc import Sequence
from datetime import UTC, datetime, timedelta
from typing import Annotated, Any
from uuid import UUID

from sqlalchemy import Select, select
from sqlalchemy.sql.operators import eq, ge, icontains_op, in_op, le

from examples.models import Author, Book, Review
from sqla_filter.base import BaseFilter
from sqla_filter.filter_ import (
    UNSET,
    FilterField,
    RelationshipInfo,
    Unset,
)


class BookFilter(BaseFilter):
    ident: Annotated[UUID | Unset, FilterField(Book.id, operator=eq)] = UNSET
    created_at_from: Annotated[
        datetime | Unset,
        FilterField(Book.created_at, operator=ge),
    ] = UNSET
    created_at_to: Annotated[
        datetime | Unset,
        FilterField(Book.created_at, operator=le),
    ] = UNSET

    author_ids: Annotated[
        Sequence[UUID] | Unset,
        FilterField(
            Author.id,
            operator=in_op,
            relationship=RelationshipInfo(field=Book.authors),
            on_apply=lambda stmt, value: stmt.where(Author.id.is_not(None)),
        ),
    ] = UNSET
    review_ids: Annotated[
        list[UUID] | Unset,
        FilterField(
            Review.id,
            operator=in_op,
            relationship=RelationshipInfo(field=Book.reviews),
        ),
    ] = UNSET
    review_content_contains: Annotated[
        str | Unset,
        FilterField(
            Review.content,
            operator=icontains_op,
            relationship=RelationshipInfo(field=Book.reviews),
        ),
    ] = UNSET


def main() -> None:
    stmt = select(Book)
    print_stmt(stmt)

    now = datetime.now(tz=UTC)
    filter_ = BookFilter(
        ident=uuid.uuid4(),
        author_ids=[uuid.uuid4()],
        review_content_contains="Review Content",
        created_at_from=now,
        created_at_to=now + timedelta(days=1),
    )

    stmt = filter_.apply(stmt)
    print_stmt(stmt)


def print_stmt(stmt: Select[tuple[Any, ...]]) -> None:
    from sqlalchemy.dialects import postgresql

    print(  # noqa: T201
        stmt.compile(
            dialect=postgresql.dialect(),  # type: ignore[no-untyped-call]
            compile_kwargs={
                "literal_binds": True,
            },
        ),
        "\n",
    )


if __name__ == "__main__":
    main()



import dataclasses
import enum
from typing import Any, Protocol, TypeAlias, TypeVar

from sqlalchemy import ColumnExpressionArgument, Select
from sqlalchemy.orm import DeclarativeBase, InstrumentedAttribute
from sqlalchemy.sql.operators import OperatorType
from sqlalchemy.sql.roles import OnClauseRole

_OnClauseArgument: TypeAlias = ColumnExpressionArgument[Any] | OnClauseRole


class Unset(enum.Enum):
    v = enum.auto()


UNSET = Unset.v

_T = TypeVar("_T", bound=DeclarativeBase)


class OnApplyHandler(Protocol):
    def __call__(self, stmt: Select[tuple[_T]], *, value: Any) -> Select[tuple[_T]]: ...


@dataclasses.dataclass(frozen=True, slots=True)
class RelationshipInfo:
    field: InstrumentedAttribute[Any]
    _: dataclasses.KW_ONLY
    onclause: _OnClauseArgument | None = None
    isouter: bool = False
    full: bool = False


@dataclasses.dataclass(frozen=True, slots=True)
class FilterField:
    field: InstrumentedAttribute[Any]
    _: dataclasses.KW_ONLY
    operator: OperatorType
    relationship: RelationshipInfo | None = None
    on_apply: OnApplyHandler | None = None



import uuid
from datetime import datetime

from sqlalchemy import ForeignKey, MetaData
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column, relationship

meta = MetaData(
    naming_convention={
        "ix": "ix_%(column_0_label)s",
        "uq": "uq_%(table_name)s_%(column_0_name)s",
        "ck": "ck_%(table_name)s_%(constraint_name)s",
        "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
        "pk": "pk_%(table_name)s",
    },
)


class Base(DeclarativeBase):
    metadata = meta


class User(Base):
    __tablename__ = "user"

    id: Mapped[uuid.UUID] = mapped_column(primary_key=True, default=uuid.uuid4)


class Review(Base):
    __tablename__ = "review"

    id: Mapped[uuid.UUID] = mapped_column(primary_key=True, default=uuid.uuid4)
    content: Mapped[str]
    book_id: Mapped[uuid.UUID] = mapped_column(ForeignKey("book.id"))
    user_id: Mapped[uuid.UUID] = mapped_column(ForeignKey("user.id"))


class Author(Base):
    __tablename__ = "author"

    id: Mapped[uuid.UUID] = mapped_column(primary_key=True, default=uuid.uuid4)


class BookAuthor(Base):
    __tablename__ = "book__author"

    book_id: Mapped[uuid.UUID] = mapped_column(ForeignKey("book.id"), primary_key=True)
    author_id: Mapped[uuid.UUID] = mapped_column(
        ForeignKey("author.id"),
        primary_key=True,
    )


class Book(Base):
    __tablename__ = "book"

    id: Mapped[uuid.UUID] = mapped_column(primary_key=True, default=uuid.uuid4)
    authors: Mapped[list[Author]] = relationship(secondary=BookAuthor.__table__)
    reviews: Mapped[list[Review]] = relationship()
    created_at: Mapped[datetime]



from .base import BaseFilter
from .filter_ import (
    UNSET,
    FilterField,
    RelationshipInfo,
    Unset,
)

__all__ = [
    "Unset",
    "UNSET",
    "RelationshipInfo",
    "FilterField",
    "BaseFilter",
]



import contextlib
import dataclasses
from collections.abc import Mapping
from typing import (
    Any,
    ClassVar,
    TypeVar,
    dataclass_transform,
    get_args,
    get_type_hints,
)

from sqlalchemy import Select
from sqlalchemy.orm import DeclarativeBase

from .filter_ import FilterField, RelationshipInfo, Unset

_TModel = TypeVar("_TModel", bound=DeclarativeBase)


@dataclass_transform(kw_only_default=True)
class BaseFilter:
    __sqla_filter_fields__: ClassVar[Mapping[str, FilterField]]

    def __init_subclass__(cls) -> None:
        wrapped_cls = dataclasses.dataclass(cls)
        class_type_hints = get_type_hints(wrapped_cls, include_extras=True)
        del class_type_hints["__sqla_filter_fields__"]

        filter_fields = {}

        for field_name, raw_type in class_type_hints.items():
            args = get_args(raw_type)
            filter_field = _get_filter_field(args)
            if filter_field is None:
                continue

            filter_fields[field_name] = filter_field

        wrapped_cls.__sqla_filter_fields__ = filter_fields

    def apply(self, stmt: Select[tuple[_TModel]]) -> Select[tuple[_TModel]]:
        for field in dataclasses.fields(self):  # type:ignore[arg-type] # mypy moment
            if (value := getattr(self, field.name)) is Unset.v:
                continue

            filter_ = self.__sqla_filter_fields__[field.name]
            if filter_.relationship:
                stmt = _apply_join(stmt, relationship=filter_.relationship)

            if filter_.on_apply:
                stmt = filter_.on_apply(stmt, value=value)

            stmt = stmt.where(filter_.operator(filter_.field, value))

        return stmt


def _get_filter_field(
    annotations: tuple[Any, ...],
) -> FilterField | None:
    with contextlib.suppress(IndexError):
        if not isinstance((filter_ := annotations[1]), FilterField):
            return None

        return filter_

    return None


def _apply_join(
    stmt: Select[tuple[_TModel]],
    *,
    relationship: RelationshipInfo,
) -> Select[tuple[_TModel]]:
    return stmt.join(
        relationship.field,
        relationship.onclause,
        isouter=relationship.isouter,
        full=relationship.full,
    )



import uuid

from sqlalchemy import select

from tests.sqla_filter.common.filter import BookFilter
from tests.sqla_filter.common.models import Author, Book, Review
from tests.utils import compile_stmt


def test_build_simple_stmt() -> None:
    ident = uuid.uuid4()

    stmt = select(Book)
    filter_ = BookFilter(ident=ident)
    stmt = filter_.apply(stmt)

    expected_stmt = select(Book).where(Book.id == ident)

    compiled_stmt = compile_stmt(stmt)
    compiled_expected_stmt = compile_stmt(expected_stmt)

    assert compiled_stmt.string == compiled_expected_stmt.string


def test_build_stmt_with_join() -> None:
    ident = uuid.uuid4()
    author_ids = [uuid.uuid4() for _ in range(3)]
    review_ids = [uuid.uuid4() for _ in range(3)]

    stmt = select(Book)
    filter_ = BookFilter(ident=ident, author_ids=author_ids, review_ids=review_ids)
    stmt = filter_.apply(stmt)

    expected_stmt = (
        select(Book)
        .where(
            Book.id == ident,
            Author.id.in_(author_ids),
            Review.id.in_(review_ids),
        )
        .join(Book.authors)
        .join(Book.reviews)
    )

    compiled_stmt = compile_stmt(stmt)
    compiled_expected_stmt = compile_stmt(expected_stmt)

    assert compiled_stmt.string == compiled_expected_stmt.string



from typing import Annotated
from uuid import UUID

from sqlalchemy.sql.operators import eq

from sqla_filter import UNSET, BaseFilter, FilterField, Unset
from tests.sqla_filter.common.models import Book


def test_base_filter() -> None:
    class Filter(BaseFilter):
        second_manually_filter: list[UUID]
        first_manually_filter: UUID | None = None
        manually_filter_field: UUID | None = None
        ident: Annotated[UUID | Unset, FilterField(Book.id, operator=eq)] = UNSET

    filter_ = Filter(second_manually_filter=[])

    assert filter_.ident is UNSET
    assert filter_.second_manually_filter == []
    assert filter_.first_manually_filter is None
